using System;
using Microsoft.SPOT;
using System.Net;
using System.Threading;
using System.IO;
using System.Text;
using Microsoft.SPOT.Hardware;

namespace CameraViewer
{
    public class MjpegSource : ICameraSource
    {

        private Uri uri;
        private Thread loadThread;

        public MjpegSource(Uri uri)
        {
            this.uri = uri;
        }

        #region ICameraSource Members

        public event CameraEventHandler CameraFrame;

        public void Stop()
        {
            if (loadThread != null)
            {
                loadThread.Abort();
                loadThread.Join();
            }
        }

        public void Start()
        {
            if (loadThread == null)
            {
                loadThread = new Thread(LoadFrames);
                loadThread.Start();
            }
        }

        #endregion

        private void LoadFrames()
        {
            try
            {
                WebRequest request = HttpWebRequest.Create(this.uri);
                HttpWebResponse response = (HttpWebResponse)request.GetResponse();

                if (response.StatusCode == HttpStatusCode.OK)
                {
                    LoadResponse(response);
                }
            }
            finally
            {
                loadThread = null;
            }
        }

        private void LoadResponse(HttpWebResponse response)
        {
            const string boundaryName = "boundary=";
            int b = response.ContentType.ToLower().IndexOf(boundaryName);
            if (b >= 0)
            {
                string boundary = response.ContentType.Substring(b + boundaryName.Length);
                using (Stream stream = response.GetResponseStream())
                {
                    LoadStream(boundary, stream);
                }
            }
        }

        private void LoadStream(string boundary, Stream stream)
        {
            int i = 0;
            int match = 0;
            long lastCheck = 0;
            byte[] boundaryByte = Encoding.UTF8.GetBytes(boundary + "\r\n");
            byte[] returnByte = new byte[2] { 13, 10 };
            byte[] buffer = new byte[0x2000];

            MemoryStream imageStream = new MemoryStream();

            while ((i = stream.Read(buffer, 0, buffer.Length)) > 0)
            {
                imageStream.Write(buffer, 0, i);

                // Cerco il boundary
                long p = IndexOfRange(imageStream, boundaryByte, lastCheck, out match);
                if (p >= 0)
                {
                    byte[] image = new byte[p];
                    byte[] rawImage = imageStream.ToArray();
                    // Estraggo i byte dell'immagine fino al punto in cui  stato trovato il boundary
                    Array.Copy(rawImage, 0, image, 0, image.Length);

                    if (image.Length > 0 && CameraFrame != null)
                    {
                        Debug.Print("Frame");
                        CameraFrame(new Bitmap(image, Bitmap.BitmapImageType.Jpeg));
                    }

                    // Metto nello stream i byte rimanenti
                    imageStream.SetLength(0);

                    int contentTypeOffset = (int)IndexOfRange(rawImage, returnByte, (int)p + boundaryByte.Length, out match);
                    int contentLengthOffset = 0;
                    if (contentTypeOffset > 0)
                    {
                        contentTypeOffset += returnByte.Length;
                        contentLengthOffset = (int)IndexOfRange(rawImage, returnByte, contentTypeOffset, out match);
                        if (contentLengthOffset < 0)
                            contentLengthOffset = 0;
                        else
                        {
                            contentLengthOffset += returnByte.Length;
                            contentTypeOffset = returnByte.Length;
                        }
                    }
                    else
                        contentTypeOffset = 0;

                    imageStream.Write(rawImage, contentLengthOffset + contentTypeOffset, rawImage.Length - contentLengthOffset - contentTypeOffset);
                }

                lastCheck = imageStream.Length - boundaryByte.Length + 1;
            }


        }

        private long IndexOfRange(byte[] source, byte[] find, long offset, out int match)
        {
            using (MemoryStream stream = new MemoryStream(source))
            {
                return IndexOfRange(stream, find, offset, out match);
            }
        }


        private long IndexOfRange(Stream stream, byte[] find, long offset, out int match)
        {
            long p = stream.Position;
            stream.Position = offset;
            int findX = 0;

            while (stream.Position < stream.Length - find.Length - findX - 1)
            {
                int next = stream.ReadByte();
                if (next == find[findX])
                {
                    findX++;
                    if (findX >= find.Length)
                        break;
                }
                else
                    findX = 0;
            }

            match = findX;
            long pos = stream.Position - find.Length;
            stream.Seek(0, SeekOrigin.End);

            if (findX == find.Length)
                return pos;
            else
                return -1;
        }
    }
}
